/*
 *  Copyright (C) 2002-2005  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* $Id: directserial_win32.cpp,v 1.5 2005/02/10 10:21:08 qbix79 Exp $ */

#include "dosbox.h"

#if C_DIRECTSERIAL

/* Windows version */
#if defined (WIN32)

#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "setup.h"
#include "serialport.h"

// Win32 related headers
#include <windows.h>

/* This is a serial passthrough class.  Its amazingly simple to */
/* write now that the serial ports themselves were abstracted out */


class CDirectSerial : public CSerial {
public:
	HANDLE hCom;
	DCB dcb;
	BOOL fSuccess;

	CDirectSerial(	char * realPort,
					Bit16u baseAddr,
					Bit8u initIrq,
					Bit32u initBps,
					Bit16u bytesize,
					char *parity,
					Bit16u stopbits
		) : CSerial(baseAddr,
					initIrq,
					initBps)
	{
		
		LOG_MSG("Opening Windows serial port");
		hCom = CreateFile(
					realPort,
					GENERIC_READ | GENERIC_WRITE,
                    0,					// must be opened with exclusive-access
                    NULL,				// no security attributes
                    OPEN_EXISTING,		// must use OPEN_EXISTING
                    0,					// non overlapped I/O
                    NULL				// hTemplate must be NULL for comm devices
                    );

		if (hCom == INVALID_HANDLE_VALUE) 
		{
			LOG_MSG("CreateFile failed with error %d.\n", GetLastError());
			hCom = 0;
			return;
		}

		fSuccess = GetCommState(hCom, &dcb);

		if (!fSuccess) 
		{
			// Handle the error.
			LOG_MSG("GetCommState failed with error %d.\n", GetLastError());
			return;
		}


		dcb.BaudRate = initBps;     // set the baud rate
		dcb.ByteSize = bytesize;             // data size, xmit, and rcv
		if(parity[0] == 'N') 
			dcb.Parity = NOPARITY;        // no parity bit
		if(parity[1] == 'E')
			dcb.Parity = EVENPARITY;        // even parity bit
		if(parity[2] == 'O')
			dcb.Parity = ODDPARITY;        // odd parity bit


		if(stopbits == 1) 
			dcb.StopBits = ONESTOPBIT;    // one stop bit
		if(stopbits == 2)
			dcb.StopBits = TWOSTOPBITS;   // two stop bits

		fSuccess = SetCommState(hCom, &dcb);

		// Configure timeouts to effectively use polling
		COMMTIMEOUTS ct;
		ct.ReadIntervalTimeout = MAXDWORD;
		ct.ReadTotalTimeoutConstant = 0;
		ct.ReadTotalTimeoutMultiplier = 0;
		ct.WriteTotalTimeoutConstant = 0;
		ct.WriteTotalTimeoutMultiplier = 0;
		SetCommTimeouts(hCom, &ct);

	}

	~CDirectSerial() {
		if(hCom != INVALID_HANDLE_VALUE) CloseHandle(hCom);
	}

	bool CanRecv(void) { return true; }
	bool CanSend(void) { return true; }

//	void Send(Bit8u val) { tqueue->addb(val); }

//	Bit8u Recv(Bit8u val) { return rqueue->getb(); }

	
/*****************************************************************************/
/* updatePortConfig is called when emulated app changes the serial port		**/
/* parameters baudrate, stopbits, number of databits, parity. 				**/
/*****************************************************************************/
	void updatePortConfig(Bit8u dll, Bit8u dlm, Bit8u lcr)
	{
		Bit8u parity=0;
		Bit8u bytelength=0;
		Bit16u baudrate = 0;
		
		// baud
		baudrate = dlm;
		baudrate = baudrate << 8;
		baudrate|=dll;
		if(baudrate<=0x1)			dcb.BaudRate=CBR_115200;
		else if(baudrate<=0x2)		dcb.BaudRate=CBR_57600;
		else if(baudrate<=0x3)		dcb.BaudRate=CBR_38400;
		else if(baudrate<=0x6)		dcb.BaudRate=CBR_19200;
		else if(baudrate<=0xc)		dcb.BaudRate=CBR_9600;
		else if(baudrate<=0x18)		dcb.BaudRate=CBR_4800;
		else if(baudrate<=0x30)		dcb.BaudRate=CBR_2400;
		else if(baudrate<=0x60)		dcb.BaudRate=CBR_1200;
		else if(baudrate<=0xc0)		dcb.BaudRate=CBR_600;
		else if(baudrate<=0x180)	dcb.BaudRate=CBR_300;
		else dcb.BaudRate=CBR_110;

		// byte length
		bytelength = lcr&0x3;
		bytelength+=5;
		dcb.ByteSize=bytelength;

		// parity
		parity=lcr&0x38;
		parity = parity >> 3;
		switch(parity)
		{
		case 0x1:
			{
				dcb.Parity=ODDPARITY;
				break;
			}
		case 0x3:
			{
				dcb.Parity=EVENPARITY;
				break;
			}
		case 0x5:
			{
				dcb.Parity=MARKPARITY;
				break;
			}
		case 0x7:
			{
				dcb.Parity=SPACEPARITY;
				break;
			}
		default:
			{
				dcb.Parity=NOPARITY;
				break;
			}
		}

		// stopbits
		if(lcr&0x4)
		{
			if(bytelength==5) dcb.StopBits=ONE5STOPBITS;
			else dcb.StopBits=TWOSTOPBITS;
		}
		else
		{
			dcb.StopBits=ONESTOPBIT;
		}

		SetCommState(hCom, &dcb);
		//LOG_MSG("Serial port at 0x%x: Port params changed: %d Baud", base,dcb.BaudRate);

	}

	void updateMSR()
	{
		Bit8u newmsr = 0;
		DWORD dptr;

		if(!GetCommModemStatus(hCom,&dptr))
		{
			//LOG_MSG("Serial port at %x: GetCommModemStatus failed!", base);
		}
		if(dptr&MS_CTS_ON) newmsr|=MSR_CTS_MASK;
		if(dptr&MS_DSR_ON) newmsr|=MSR_DSR_MASK;
		if(dptr&MS_RING_ON) newmsr|=MSR_RI_MASK;
		if(dptr&MS_RLSD_ON) newmsr|=MSR_CD_MASK;
		changeMSR(newmsr);

	}



	void transmitByte(Bit8u val)
	{
		DWORD bytesWritten;
		WriteFile(hCom, &val, 1, &bytesWritten, NULL);
		if(&bytesWritten >0)
		{
			ByteTransmitted();
			//LOG_MSG("UART 0x%x: TX 0x%x", base,val);
		}
		else
		{
			LOG_MSG("UART 0x%x: NO BYTE WRITTEN! PORT HANGS NOW!", base);
		}
	}

/*****************************************************************************/
/* toggleBreak(val) switches break on or off								**/
/*****************************************************************************/

	void toggleBreak(bool value)
	{
		LOG_MSG("UART 0x%x: Break toggeled: %d", base, value);
		if(value) SetCommBreak(hCom);
		else ClearCommBreak(hCom);


	}

/*****************************************************************************/
/* updateModemControlLines(mcr) sets DTR and RTS.							**/
/*****************************************************************************/
	void updateModemControlLines(Bit8u mcr)
	{
		bool change = false;
		
		/*** DTR ***/
		if(mcr&MCR_DTR_MASK)	// DTR on
		{		
			if(dcb.fDtrControl==DTR_CONTROL_DISABLE)
			{
				dcb.fDtrControl=DTR_CONTROL_ENABLE;
				change=true;
			}
		}
		else
		{
			if(dcb.fDtrControl==DTR_CONTROL_ENABLE)
			{
				dcb.fDtrControl=DTR_CONTROL_DISABLE;
				change=true;
			}
		}
		/*** RTS ***/
		if(mcr&MCR_RTS_MASK)	// RTS on
		{		
			if(dcb.fRtsControl==RTS_CONTROL_DISABLE)
			{
				dcb.fRtsControl=RTS_CONTROL_ENABLE;
				change=true;
			}
		}
		else
		{
			if(dcb.fRtsControl==RTS_CONTROL_ENABLE)
			{
				dcb.fRtsControl=RTS_CONTROL_DISABLE;
				change=true;
			}
		}
		if(change)SetCommState(hCom, &dcb);
	}
	
	void Timer2(void)
	{	
		DWORD dwRead;
		Bit8u chRead;
		// update RX
		if (ReadFile(hCom, &chRead, 1, &dwRead, NULL)) {
			if(dwRead != 0)
			{	
				//LOG_MSG("UART 0x%x: RX 0x%x", base,chRead);
				receiveByte(chRead);
				
			}
		}

		// update Modem input line states
		updateMSR();

	}

		
};

CDirectSerial *cds;

void DIRECTSERIAL_Init(Section* sec) {

	unsigned long args = 1;
	Section_prop * section=static_cast<Section_prop *>(sec);
	

	if(!section->Get_bool("directserial")) return;

	Bit16u comport = section->Get_int("comport");
	Bit32u bps = section->Get_int("defaultbps");
	switch (comport) {
		case 1:
			cds = new CDirectSerial((char *)section->Get_string("realport"), 0x3f8, 4, bps, section->Get_int("bytesize"), (char *)section->Get_string("parity"), section->Get_int("stopbit"));
			break;
		case 2:
			cds = new CDirectSerial((char *)section->Get_string("realport"), 0x2f8, 3, bps, section->Get_int("bytesize"), (char *)section->Get_string("parity"), section->Get_int("stopbit"));
			break;
		case 3:
			cds = new CDirectSerial((char *)section->Get_string("realport"), 0x3e8, 4, bps, section->Get_int("bytesize"), (char *)section->Get_string("parity"), section->Get_int("stopbit"));
			break;
		case 4:
			cds = new CDirectSerial((char *)section->Get_string("realport"), 0x2e8, 3, bps, section->Get_int("bytesize"), (char *)section->Get_string("parity"), section->Get_int("stopbit"));
			break;
		default:
			cds = new CDirectSerial((char *)section->Get_string("realport"), 0x3f8, 4, bps, section->Get_int("bytesize"), (char *)section->Get_string("parity"), section->Get_int("stopbit"));
			break;

	}



	seriallist.push_back(cds);
}
#else /*linux and others oneday maybe */

#endif 
#endif

